# pylint: disable=W0703
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import shutil

import octobot_commons.databases.run_databases.abstract_run_databases_pruner as abstract_run_databases_pruner


class FileSystemRunDatabasesPruner(
    abstract_run_databases_pruner.AbstractRunDatabasesPruner
):
    async def _explore_databases(self):
        self.all_db_data = [
            abstract_run_databases_pruner.DBData(
                directory,
                [FileSystemDBPartData(f) for f in self._get_all_files(directory)],
            )
            for directory in self._get_file_system_runs(self.databases_root_identifier)
        ]

    async def _prune_database(self, db_data):
        try:
            shutil.rmtree(db_data.identifier)
            return True
        except Exception as err:
            self.logger.exception(err, True, f"Error when deleting run database: {err}")
            return False

    async def _get_global_runs_identifiers(self, removed_databases):
        return {
            os.path.dirname(removed_database.identifier)
            for removed_database in removed_databases
        }

    def _get_file_system_runs(self, root):
        try:
            # use os.scandir as it is much faster than os.walk
            for entry in os.scandir(root):
                if self._is_run_top_level_folder(entry):
                    yield entry
                elif entry.is_dir():
                    yield from self._get_file_system_runs(entry)
        except FileNotFoundError:
            # nothing to explore
            pass

    def _get_all_files(self, root):
        for entry in os.scandir(root):
            if entry.is_file():
                yield entry
            elif entry.is_dir():
                yield from self._get_all_files(entry)

    def _is_run_top_level_folder(self, dir_entry):
        return os.path.isfile(os.path.join(dir_entry, self._run_db)) and any(
            identifier in dir_entry.path
            for identifier in self.backtesting_run_path_identifier
        )


class FileSystemDBPartData(abstract_run_databases_pruner.AbstractDBPartData):
    def __init__(self, identifier):
        super().__init__(identifier)
        self.size = os.path.getsize(self.identifier)
        self.last_modified_time = os.path.getmtime(self.identifier)



# pylint: disable=W0703,R0902,R0913,R1729
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import time

import octobot_commons.logging as logging_util
import octobot_commons.html_util as html_util


class AsyncJob:
    """
    Async job management
    """

    NO_DELAY = 0.1
    DEPENDENCIES_WAIT_TIMEOUT = 300
    SELF_RUNNING_WAIT_TIMEOUT = 30
    MAXIMUM_ALLOWED_SUCCESSIVE_FAILURES = 1

    def __init__(
        self,
        callback,
        execution_interval_delay=NO_DELAY,
        min_execution_delay=NO_DELAY,
        first_execution_delay=NO_DELAY,
        is_periodic=True,
        enable_multiple_runs=False,
        max_successive_failures=MAXIMUM_ALLOWED_SUCCESSIVE_FAILURES,
    ):
        self.logger = logging_util.get_logger(
            f"{self.__class__.__name__}-{callback.__name__}"
        )
        self.callback = callback
        self.is_started = False
        self.should_stop = False
        self.is_periodic = is_periodic
        self.enable_multiple_runs = enable_multiple_runs
        self.simultaneous_calls = 0
        self.successive_failures = 0

        # Set this attribute to 0 to log on any periodic refresh exception.
        self.max_successive_failures = max_successive_failures

        self.last_execution_time = 0
        self.execution_interval_delay = execution_interval_delay
        self.min_execution_delay = min_execution_delay
        self.first_execution_delay = first_execution_delay

        self.job_dependencies = []
        self.idle_task_event = asyncio.Event()
        self.idle_task_event.set()

        self.job_task = None
        self.job_periodic_task = None

    async def run(
        self,
        force=False,
        wait_for_task_execution=False,
        ignore_dependencies_check=False,
        **kwargs,
    ):
        """
        Run the job if possible
        Reschedule the job in the end
        :param force: When True, force the execution of the job
        :param wait_for_task_execution: When True, await idle_task_event
        :param ignore_dependencies_check: When True, ignore dependencies wait
        """
        if not self.is_started and self.is_periodic:
            self.should_stop = False
            self.job_periodic_task = asyncio.create_task(
                self._run_periodic_task(**kwargs)
            )
        else:
            if self._should_run_job(force=force, ignore_dependencies=True):
                if wait_for_task_execution:
                    await self._run_task_as_soon_as_possible(
                        force=force,
                        ignore_dependencies_check=ignore_dependencies_check,
                        **kwargs,
                    )
                else:
                    self.job_task = asyncio.create_task(
                        self._run_task_as_soon_as_possible(
                            force=force,
                            ignore_dependencies_check=ignore_dependencies_check,
                            **kwargs,
                        )
                    )

    async def _run_periodic_task(self, **kwargs):
        """
        Calls _run() periodically until self.should_stop == True or cancellation
        """
        while not self.should_stop:
            self.is_started = True
            if self.last_execution_time == 0:
                # first execution
                sleep_time = (
                    0
                    if self.first_execution_delay == self.NO_DELAY
                    else self.first_execution_delay
                )
            else:
                # other executions
                sleep_time = (
                    0
                    if time.time() - self.last_execution_time
                    >= self.execution_interval_delay
                    else self.execution_interval_delay
                )
            await asyncio.sleep(sleep_time)
            await self._run_task_as_soon_as_possible(
                error_on_single_failure=False, **kwargs
            )
        self.is_started = False

    async def _run_task_as_soon_as_possible(
        self,
        force=False,
        ignore_dependencies_check=False,
        error_on_single_failure=True,
        **kwargs,
    ):
        """
        Wait until job _run() can be called
        :param force: if True, force job task execution
        """
        if self._should_run_job(force=force):
            await self._run(error_on_single_failure=error_on_single_failure, **kwargs)
        else:
            # wait for job dependencies to stop running
            # and also this job to stop running
            try:
                events_to_wait = []

                # add dependencies event wait_for coroutines
                if not ignore_dependencies_check:
                    events_to_wait += [
                        asyncio.wait_for(
                            dependency.idle_task_event.wait(),
                            self.DEPENDENCIES_WAIT_TIMEOUT,
                        )
                        for dependency in self.job_dependencies
                    ]

                # add self idle event wait_for coroutine
                if not self.enable_multiple_runs:
                    events_to_wait.append(
                        asyncio.wait_for(
                            self.idle_task_event.wait(),
                            self.SELF_RUNNING_WAIT_TIMEOUT,
                        )
                    )

                if events_to_wait:
                    await asyncio.gather(
                        *events_to_wait,
                    )
            except asyncio.TimeoutError:
                self.logger.warning("Job has been timed out")
            finally:
                await self._run(
                    error_on_single_failure=error_on_single_failure, **kwargs
                )

    def is_job_idle(self):
        """
        :return: publicly is_running attribute value
        """
        return self.idle_task_event.is_set()

    def add_job_dependency(self, job):
        """
        Add a new job dependency
        :param job: the new job dependency
        """
        self.job_dependencies.append(job)

    async def _run(self, error_on_single_failure=True, **kwargs):
        """
        Execute the job callback
        Reset the last_execution_time
        """
        # Clear to be able to await the event
        if self.simultaneous_calls == 0:
            self.idle_task_event.clear()
        self.simultaneous_calls += 1
        try:
            await self.callback(**kwargs)
            if self.successive_failures > self.max_successive_failures:
                self.logger.info(
                    f"Job successfully run after {self.successive_failures} failures."
                )
            self.successive_failures = 0
        except Exception as exception:
            self._handle_run_exception(exception, error_on_single_failure)
        finally:
            self.last_execution_time = time.time()
            self.simultaneous_calls -= 1
            if self.simultaneous_calls == 0:
                # Set the event to trigger event waiters
                self.idle_task_event.set()

    def _handle_run_exception(self, exception, error_on_single_failure):
        self.successive_failures += 1
        str_error = html_util.get_html_summary_if_relevant(exception)
        error_message = f"Failed to run job action, exception: {exception.__class__.__name__}: {str_error}"
        if error_on_single_failure:
            self.logger.exception(exception, True, error_message)
        else:
            if self.successive_failures > self.max_successive_failures:
                self.logger.exception(
                    exception,
                    True,
                    f"{error_message} ({self.successive_failures} failures in a row)",
                )
            else:
                self.logger.debug(error_message)
                # always at least print stacktrace in logs
                self.logger.exception(exception, False)

    def _should_run_job(self, force=False, ignore_dependencies=False):
        """
        :param force: If True, enabled job execution even if min_execution_delay > last_execution_time
        :param ignore_dependencies: If True, ignore _are_job_dependencies_running() result
        :return: True if the job is not already running and if _should_run is True
        """
        return (self.is_job_idle() or self.enable_multiple_runs) and (
            (self._are_job_dependencies_idle() or ignore_dependencies)
            and (self._has_enough_time_elapsed() or force)
        )

    def _has_enough_time_elapsed(self):
        """
        :return: True if min_execution_delay < last_execution_time
        """
        return (
            time.time() - self.min_execution_delay > self.last_execution_time
            or self.min_execution_delay == AsyncJob.NO_DELAY
            or self.last_execution_time == 0
        )

    def _are_job_dependencies_idle(self):
        """
        :return: True if a dependent jobs is idle
        """
        return all([job.is_job_idle() for job in self.job_dependencies])

    def is_stopped(self):
        """
        Return True when the AsyncJob has stopped
        """
        return self.should_stop

    def stop(self):
        """
        Stop the job by cancelling the execution task
        """
        self.should_stop = True
        if self.job_task is not None:
            self.job_task.cancel()
            self.job_task = None
        if self.job_periodic_task is not None:
            self.job_periodic_task.cancel()
            self.job_periodic_task = None
        self.is_started = False

    def clear(self):
        """
        Clear job object references and stop it
        """
        self.job_dependencies = []
        self.stop()



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_commons.tentacles_management.class_inspector import default_parent_inspection, default_parents_inspection, \
    get_class_from_parent_subclasses, get_deep_class_from_parent_subclasses, \
    is_abstract_using_inspection_and_class_naming, get_all_classes_from_parent, get_single_deepest_child_class


class AbstractParent:
    pass


class Parent(AbstractParent):
    pass


class BasicChild(Parent):
    pass


class ChildOfChild(BasicChild):
    pass


def test_default_parent_inspection():
    assert default_parent_inspection(BasicChild, Parent)
    assert not default_parent_inspection(ChildOfChild, Parent)
    assert default_parent_inspection(ChildOfChild, BasicChild)
    assert not default_parent_inspection(BasicChild, ChildOfChild)


def test_default_parents_inspection():
    assert default_parents_inspection(BasicChild, Parent)
    assert default_parents_inspection(ChildOfChild, Parent)
    assert default_parents_inspection(ChildOfChild, BasicChild)
    assert not default_parents_inspection(BasicChild, ChildOfChild)


def test_get_class_from_parent_subclasses():
    assert get_class_from_parent_subclasses("BasicChild", Parent) is BasicChild
    assert get_class_from_parent_subclasses("ChildOfChild", Parent) is None
    assert get_class_from_parent_subclasses("ChildOfChild", BasicChild) is ChildOfChild
    assert get_class_from_parent_subclasses("BasicChild", ChildOfChild) is None


def test_get_deep_class_from_parent_subclasses():
    assert get_deep_class_from_parent_subclasses("BasicChild", Parent) is BasicChild
    assert get_deep_class_from_parent_subclasses("ChildOfChild", Parent) is ChildOfChild
    assert get_deep_class_from_parent_subclasses("ChildOfChild", BasicChild) is ChildOfChild
    assert get_deep_class_from_parent_subclasses("BasicChild", ChildOfChild) is None


def test_is_abstract_using_inspection_and_class_naming():
    assert is_abstract_using_inspection_and_class_naming(AbstractParent)
    assert not is_abstract_using_inspection_and_class_naming(Parent)
    assert not is_abstract_using_inspection_and_class_naming(ChildOfChild)


def test_get_all_classes_from_parent():
    assert get_all_classes_from_parent(Parent) == [BasicChild, ChildOfChild]
    assert get_all_classes_from_parent(BasicChild) == [ChildOfChild]
    assert get_all_classes_from_parent(ChildOfChild) == []


def test_get_single_deepest_child_class():
    assert get_single_deepest_child_class(Parent) == ChildOfChild
    assert get_single_deepest_child_class(BasicChild) == ChildOfChild
    assert get_single_deepest_child_class(ChildOfChild) == ChildOfChild



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import contextlib

import octobot_commons.databases.relational_databases.sqlite.cursor_wrapper as cursor_wrapper


class CursorPool:
    def __init__(self, db_connection):
        self._db_connection = db_connection
        self._cursors = []

    @contextlib.asynccontextmanager
    async def idle_cursor(self) -> cursor_wrapper.CursorWrapper:
        """
        Yields an idle cursor, creates a new one if necessary
        """
        cursor = None
        try:
            cursor = await self._get_or_create_idle_cursor()
            cursor.idle = False
            yield cursor
        finally:
            if cursor is not None:
                cursor.idle = True

    async def close(self):
        """
        Close every cursor
        """
        await asyncio.gather(*(cursor.close() for cursor in self._cursors))

    async def _get_or_create_idle_cursor(self) -> cursor_wrapper.CursorWrapper:
        for cursor in self._cursors:
            if cursor.idle:
                return cursor
        cursor = cursor_wrapper.CursorWrapper(await self._db_connection.cursor())
        self._cursors.append(cursor)
        return cursor



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_commons.tree import BaseTree, NodeExistsError


def test_base_tree_init():
    assert BaseTree()


def test_base_tree_get_new_node():
    base_tree = BaseTree()
    created_node = base_tree.get_or_create_node(["test"])
    assert base_tree.root.children == {"test": created_node}


def test_base_tree_get_existing_node():
    base_tree = BaseTree()
    created_node = base_tree.get_or_create_node(["test"])
    get_node_result = base_tree.get_or_create_node(["test"])
    assert created_node is get_node_result


def test_base_tree_get_not_existing_node():
    base_tree = BaseTree()
    with pytest.raises(NodeExistsError):
        assert base_tree.get_node(["test"]) is None


def test_base_tree_delete_existing_node():
    base_tree = BaseTree()
    created_node = base_tree.get_or_create_node(["test"])
    delete_node_result = base_tree.delete_node(["test"])
    assert created_node is delete_node_result
    with pytest.raises(NodeExistsError):
        base_tree.get_node(["test"])


def test_base_tree_delete_not_existing_node():
    base_tree = BaseTree()
    with pytest.raises(NodeExistsError):
        base_tree.delete_node(["test"])


def test_base_tree_get_new_relative_node():
    base_tree = BaseTree()
    created_node = base_tree.get_or_create_node(["test"])
    relative_created_node = base_tree.get_or_create_node(["test-relative"], starting_node=created_node)
    get_node_result = base_tree.get_or_create_node(["test", "test-relative"])
    assert relative_created_node is get_node_result


def test_base_tree_get_relative_node():
    base_tree = BaseTree()
    created_node = base_tree.get_or_create_node(["test"])
    relative_created_node = base_tree.get_or_create_node(["test", "test-relative"])
    get_node_result = base_tree.get_or_create_node(["test-relative"], starting_node=created_node)
    assert relative_created_node is get_node_result


def test_base_tree_delete_relative_node():
    base_tree = BaseTree()
    created_node = base_tree.get_or_create_node(["test"])
    relative_created_node = base_tree.get_or_create_node(["test", "test-relative"])
    delete_node_result = base_tree.delete_node(["test-relative"], starting_node=created_node)
    assert relative_created_node is delete_node_result
    assert base_tree.get_node(["test"]) is created_node
    with pytest.raises(NodeExistsError):
        base_tree.get_node(["test", "test-relative"])


def test_base_tree_set_node():
    base_tree = BaseTree()
    created_node = base_tree.get_or_create_node(["test"])
    base_tree.set_node(1, None, created_node)
    assert created_node.node_value == 1
    assert created_node.node_type is None
    base_tree.set_node(5, None, created_node, timestamp=10)
    assert created_node.node_value == 5
    assert created_node.node_type is None
    assert created_node.node_value_time == 10


def test_base_tree_set_node_at_path():
    base_tree = BaseTree()
    base_tree.set_node_at_path("test-string", "test-type", ["test", "test2", "test3"])
    assert base_tree.get_or_create_node(["test"])
    assert base_tree.get_or_create_node(["test", "test2"])
    assert base_tree.get_or_create_node(["test", "test2", "test3"])
    assert base_tree.get_or_create_node(["test"]).children
    assert base_tree.get_or_create_node(["test", "test2"]).children
    assert not base_tree.get_or_create_node(["test", "test2", "test3"]).children
    assert base_tree.get_or_create_node(["test", "test2", "test3"]).node_value == "test-string"
    assert base_tree.get_or_create_node(["test", "test2", "test3"]).node_type == "test-type"


def test_get_children_keys():
    base_tree = BaseTree()
    base_tree.set_node_at_path("test-string", "test-type", ["test", "test2", "test3"])
    base_tree.set_node_at_path("test-string_2", None, ["test", "test2", "test3_2"])
    base_tree.set_node_at_path("test-string_3", None, ["test", "test2"])
    base_tree.set_node_at_path("test-string_4", None, ["test", "test3"])
    assert base_tree.get_children_keys([]) == ["test"]
    assert base_tree.get_children_keys(["test"]) == ["test2", "test3"]
    assert base_tree.get_children_keys(["test", "test2"]) == ["test3", "test3_2"]
    with pytest.raises(NodeExistsError):
        assert base_tree.get_children_keys(["test", "testXXXX"]) == ["test3"]


def test_get_nested_children_with_path():
    base_tree = BaseTree()
    base_tree.set_node_at_path("test-string", "test-type", ["test", "test2", "test3"])
    base_tree.set_node_at_path("test-string_2", None, ["test", "test2", "test3_2"])
    base_tree.set_node_at_path("test-string_3", None, ["test", "test2"])
    base_tree.set_node_at_path("test-string_4", None, ["test", "test3"])
    assert [(n.node_value, p) for n, p in base_tree.get_nested_children_with_path()] == [
        ("test-string", ["test", "test2", "test3"]),
        ("test-string_2", ["test", "test2", "test3_2"]),
        ("test-string_4", ["test", "test3"])
    ]
    assert [(n.node_value, p) for n, p in base_tree.get_nested_children_with_path(select_leaves_only=False)] == [
        (None, []),
        (None, ['test']),
        ("test-string_3", ["test", "test2"]),
        ("test-string", ["test", "test2", "test3"]),
        ("test-string_2", ["test", "test2", "test3_2"]),
        ("test-string_4", ["test", "test3"])
    ]
    assert [(n.node_value, p) for n, p in base_tree.get_nested_children_with_path(path=["test", "test2"],
                                                                                   select_leaves_only=False)] == [
        ("test-string_3", ["test", "test2"]),
        ("test-string", ["test", "test2", "test3"]),
        ("test-string_2", ["test", "test2", "test3_2"])
    ]



# pylint: disable=C0116
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.signals.signal as signal
import octobot_commons.signals.signal_bundle as signal_bundle
import octobot_commons.enums as commons_enums


def create_signal_bundle(signal_bundle_dict: dict) -> signal_bundle.SignalBundle:
    signal_bundle_value = signal_bundle_dict[
        commons_enums.CommunityFeedAttrs.VALUE.value
    ]
    return signal_bundle.SignalBundle(
        signal_bundle_value.get(commons_enums.SignalBundlesAttrs.IDENTIFIER.value),
        signals=[
            create_signal(s)
            for s in signal_bundle_value.get(
                commons_enums.SignalBundlesAttrs.SIGNALS.value, []
            )
        ],
        version=signal_bundle_value.get(commons_enums.SignalBundlesAttrs.VERSION.value),
    )


def create_signal(signal_dict: dict) -> signal.Signal:
    return signal.Signal(
        signal_dict.get(commons_enums.SignalsAttrs.TOPIC.value),
        signal_dict.get(commons_enums.SignalsAttrs.CONTENT.value),
    )
